@@ -2,6 +2,7 @@
 /*
 NetJack Driver
 
+Copyright (C) 2008 Pieter Palmers <pieterpalmers@users.sourceforge.net>
 Copyright (C) 2006 Torben Hohn <torbenh@gmx.de>
 Copyright (C) 2003 Robert Ham <rah@bash.sh>
 Copyright (C) 2001 Paul Davis
@@ -31,7 +32,6 @@ $Id: net_driver.c,v 1.17 2006/04/16 20:16:10 torbenh Exp $
 #include <errno.h>
 #include <stdarg.h>
 #include <sys/mman.h>
-#include <sys/poll.h>
 
 #include <jack/types.h>
 #include <jack/engine.h>
@@ -50,8 +50,10 @@ $Id: net_driver.c,v 1.17 2006/04/16 20:16:10 torbenh Exp $
 
 static int sync_state = TRUE;
 static jack_transport_state_t last_transport_state;
-static int framecnt; // this is set upon reading a packet.
-// and will be written into the pkthdr of an outgoing packet.
+
+/* This is set upon reading a packetand will be
+ * written into the pkthdr of an outgoing packet */
+static int framecnt; 
 
 static int
 net_driver_sync_cb(jack_transport_state_t state, jack_position_t *pos, net_driver_t *driver)
@@ -68,8 +70,7 @@ net_driver_sync_cb(jack_transport_state_t state, jack_position_t *pos, net_drive
 }
 
 static jack_nframes_t
-net_driver_wait (net_driver_t *driver, int extra_fd, int *status,
-                 float *delayed_usecs)
+net_driver_wait (net_driver_t *driver, int extra_fd, int *status, float *delayed_usecs)
 {
     // ok... we wait for a packet on the socket
     // TODO:
@@ -80,45 +81,28 @@ net_driver_wait (net_driver_t *driver, int extra_fd, int *status,
     // on packet loss we should either detect an xrun or just continue running when we
     // think, that the sync source is not running anymore.
 
-#if 0
-    fd_set read_fds;
-    int res = 0;
-    while (res == 0) {
-        FD_ZERO(&read_fds);
-        FD_SET(driver->sockfd, &read_fds);
-        res = select(driver->sockfd + 1, &read_fds, NULL, NULL, NULL); // wait here until there is a packet to get
-    }
-#endif
+    socklen_t address_size = sizeof (struct sockaddr_in);
+    int bufsize, len;
+    jacknet_packet_header *pkthdr = (jacknet_packet_header *) driver->rx_buf;
+    
+    bufsize =  get_sample_size (driver->bitdepth) * driver->capture_channels * driver->net_period_down + sizeof (jacknet_packet_header);
 
-    socklen_t address_size = sizeof(struct sockaddr_in);
-
-    int bufsize =  get_sample_size(driver->bitdepth) * driver->capture_channels * driver->net_period_down + sizeof(jacknet_packet_header);
-
-    jacknet_packet_header *pkthdr = (jacknet_packet_header *)driver->rx_buf;
-
-rx_again:
-    if (!driver->srcaddress_valid) {
-        // XXX: Somthings really bad ;)
-        puts("!driver->srcaddress_valid");
-        return 0;
-    }
-    int len = netjack_recvfrom(driver->sockfd, (char *)driver->rx_buf, bufsize,
-                                MSG_WAITALL, (struct sockaddr*) & driver->syncsource_address, &address_size, 1400);
-
-
-    if (len != bufsize) {
-        jack_info("wrong_packet_len: len=%d, expected=%d", len, bufsize);
-        goto rx_again;
+    if (netjack_poll (driver->sockfd, 500))
+        len = netjack_recvfrom (driver->sockfd, (char *)driver->rx_buf, bufsize, MSG_WAITALL, (struct sockaddr*) & driver->syncsource_address, &address_size, driver->mtu);
+    else
+        len = 0;
+    while (len != bufsize)
+    {
+        jack_error ("wrong_packet_len: len=%d, expected=%d", len, bufsize);
+        if (netjack_poll (driver->sockfd, 500))
+            len = netjack_recvfrom (driver->sockfd, (char *)driver->rx_buf, bufsize, MSG_WAITALL, (struct sockaddr*) & driver->syncsource_address, &address_size, driver->mtu);
+        else
+            len = 0;
     }
 
-    packet_header_ntoh(pkthdr);
-
-    //driver->srcaddress_valid = 0;
-
+    packet_header_ntoh (pkthdr);
     driver->last_wait_ust = jack_get_microseconds ();
-    driver->engine->transport_cycle_start (driver->engine,
-                                           driver->last_wait_ust);
-
+    driver->engine->transport_cycle_start (driver->engine, driver->last_wait_ust);
     /* this driver doesn't work so well if we report a delay */
     *delayed_usecs = 0;		/* lie about it */
     *status = 0;
@@ -129,7 +113,7 @@ static inline int
 net_driver_run_cycle (net_driver_t *driver)
 {
     jack_engine_t *engine = driver->engine;
-    int wait_status;
+    int wait_status = -1;
     float delayed_usecs;
 
     jack_nframes_t nframes = net_driver_wait (driver, -1, &wait_status,
@@ -190,7 +174,7 @@ net_driver_null_cycle (net_driver_t* driver, jack_nframes_t nframes)
             driver->syncsource_address.sin_port = htons(driver->reply_port);
 
     netjack_sendto(driver->outsockfd, (char *)packet_buf, tx_size,
-                    0, (struct sockaddr*)&driver->syncsource_address, sizeof(struct sockaddr_in), 1400);
+                    0, (struct sockaddr*)&driver->syncsource_address, sizeof(struct sockaddr_in), driver->mtu);
 
     return 0;
 }
@@ -224,8 +208,8 @@ net_driver_read (net_driver_t* driver, jack_nframes_t nframes)
     //packet_header_ntoh(pkthdr);
     // fill framecnt from pkthdr.
 
-    if (pkthdr->framecnt != framecnt + 1)
-        jack_info("bogus framecount %d", pkthdr->framecnt);
+    //if (pkthdr->framecnt != framecnt + 1)
+    //    jack_info("bogus framecount %d", pkthdr->framecnt);
 
     framecnt = pkthdr->framecnt;
     driver->reply_port = pkthdr->reply_port;
@@ -270,7 +254,7 @@ net_driver_read (net_driver_t* driver, jack_nframes_t nframes)
 //				    jack_info("running locate to %d", pkthdr->transport_frame + (pkthdr->latency)*nframes);
 //		    		}
                 if (local_trans_state != JackTransportRolling)
-                    jack_transport_start(driver->client);
+                    jack_transport_start (driver->client);
                 break;
 
             case JackTransportLooping:
@@ -278,8 +262,7 @@ net_driver_read (net_driver_t* driver, jack_nframes_t nframes)
         }
     }
 
-
-    render_payload_to_jack_ports(driver->bitdepth, packet_bufX, driver->net_period_down, driver->capture_ports, driver->capture_srcs, nframes);
+    render_payload_to_jack_ports (driver->bitdepth, packet_bufX, driver->net_period_down, driver->capture_ports, driver->capture_srcs, nframes);
 
     return 0;
 }
@@ -308,7 +291,7 @@ net_driver_write (net_driver_t* driver, jack_nframes_t nframes)
         if (driver->reply_port)
             driver->syncsource_address.sin_port = htons(driver->reply_port);
     netjack_sendto(driver->outsockfd, (char *)packet_buf, packet_size,
-                    0, (struct sockaddr*)&driver->syncsource_address, sizeof(struct sockaddr_in), 1400);
+                    0, (struct sockaddr*)&driver->syncsource_address, sizeof(struct sockaddr_in), driver->mtu);
 
     return 0;
 }
@@ -317,7 +300,7 @@ net_driver_write (net_driver_t* driver, jack_nframes_t nframes)
 static int
 net_driver_attach (net_driver_t *driver)
 {
-    puts ("net_driver_attach");
+    //puts ("net_driver_attach");
     jack_port_t * port;
     char buf[32];
     unsigned int chn;
@@ -331,14 +314,14 @@ net_driver_attach (net_driver_t *driver)
 
     port_flags = JackPortIsOutput | JackPortIsPhysical | JackPortIsTerminal;
 
-    for (chn = 0; chn < driver->capture_channels; chn++) {
+    for (chn = 0; chn < driver->capture_channels_audio; chn++) {
         snprintf (buf, sizeof(buf) - 1, "capture_%u", chn + 1);
 
         port = jack_port_register (driver->client, buf,
                                    JACK_DEFAULT_AUDIO_TYPE,
                                    port_flags, 0);
         if (!port) {
-            jack_error ("DUMMY: cannot register port for %s", buf);
+            jack_error ("NET: cannot register port for %s", buf);
             break;
         }
 
@@ -346,10 +329,25 @@ net_driver_attach (net_driver_t *driver)
             jack_slist_append (driver->capture_ports, port);
         driver->capture_srcs = jack_slist_append(driver->capture_srcs, src_new(SRC_LINEAR, 1, NULL));
     }
+    for (chn = driver->capture_channels_audio; chn < driver->capture_channels; chn++) {
+        snprintf (buf, sizeof(buf) - 1, "capture_%u", chn + 1);
+
+        port = jack_port_register (driver->client, buf,
+                                   JACK_DEFAULT_MIDI_TYPE,
+                                   port_flags, 0);
+        if (!port) {
+            jack_error ("NET: cannot register port for %s", buf);
+            break;
+        }
+
+        driver->capture_ports =
+            jack_slist_append (driver->capture_ports, port);
+        //driver->capture_srcs = jack_slist_append(driver->capture_srcs, src_new(SRC_LINEAR, 1, NULL));
+    }
 
     port_flags = JackPortIsInput | JackPortIsPhysical | JackPortIsTerminal;
 
-    for (chn = 0; chn < driver->playback_channels; chn++) {
+    for (chn = 0; chn < driver->playback_channels_audio; chn++) {
         snprintf (buf, sizeof(buf) - 1, "playback_%u", chn + 1);
 
         port = jack_port_register (driver->client, buf,
@@ -357,7 +355,7 @@ net_driver_attach (net_driver_t *driver)
                                    port_flags, 0);
 
         if (!port) {
-            jack_error ("DUMMY: cannot register port for %s", buf);
+            jack_error ("NET: cannot register port for %s", buf);
             break;
         }
 
@@ -365,6 +363,22 @@ net_driver_attach (net_driver_t *driver)
             jack_slist_append (driver->playback_ports, port);
         driver->playback_srcs = jack_slist_append(driver->playback_srcs, src_new(SRC_LINEAR, 1, NULL));
     }
+    for (chn = driver->playback_channels_audio; chn < driver->playback_channels; chn++) {
+        snprintf (buf, sizeof(buf) - 1, "playback_%u", chn + 1);
+
+        port = jack_port_register (driver->client, buf,
+                                   JACK_DEFAULT_MIDI_TYPE,
+                                   port_flags, 0);
+
+        if (!port) {
+            jack_error ("NET: cannot register port for %s", buf);
+            break;
+        }
+
+        driver->playback_ports =
+            jack_slist_append (driver->playback_ports, port);
+        //driver->playback_srcs = jack_slist_append(driver->playback_srcs, src_new(SRC_LINEAR, 1, NULL));
+    }
 
     jack_activate (driver->client);
     return 0;
@@ -408,6 +422,8 @@ net_driver_new (jack_client_t * client,
                 char *name,
                 unsigned int capture_ports,
                 unsigned int playback_ports,
+                unsigned int capture_ports_midi,
+                unsigned int playback_ports_midi,
                 jack_nframes_t sample_rate,
                 jack_nframes_t period_size,
                 unsigned int listen_port,
@@ -417,9 +433,10 @@ net_driver_new (jack_client_t * client,
                 unsigned int bitdepth)
 {
     net_driver_t * driver;
+    int first_pack_len, rx_bufsize;
     struct sockaddr_in address;
 
-    jack_info("creating net driver ... %s|%" PRIu32 "|%" PRIu32
+    jack_info ("creating net driver ... %s|%" PRIu32 "|%" PRIu32
             "|%u|%u|%u|transport_sync:%u", name, sample_rate, period_size, listen_port,
             capture_ports, playback_ports, transport_sync);
 
@@ -445,17 +462,22 @@ net_driver_new (jack_client_t * client,
     driver->listen_port   = listen_port;
     driver->last_wait_ust = 0;
 
-    driver->capture_channels  = capture_ports;
+    driver->capture_channels  = capture_ports + capture_ports_midi;
+    driver->capture_channels_audio  = capture_ports;
+    driver->capture_channels_midi   = capture_ports_midi;
     driver->capture_ports     = NULL;
-    driver->playback_channels = playback_ports;
+    driver->playback_channels = playback_ports + playback_ports_midi;
+    driver->playback_channels_audio = playback_ports;
+    driver->playback_channels_midi = playback_ports_midi;
     driver->playback_ports    = NULL;
 
     driver->handle_transport_sync = transport_sync;
     driver->client = client;
     driver->engine = NULL;
 
-    if ((bitdepth != 0) && (bitdepth != 8) && (bitdepth != 16)) {
-        jack_info("Invalid bitdepth: %d (8,16 or 0 for float) !!!", bitdepth);
+    if ((bitdepth != 0) && (bitdepth != 8) && (bitdepth != 16))
+    {
+        jack_info ("Invalid bitdepth: %d (8, 16 or 0 for float) !!!", bitdepth);
         return NULL;
     }
     driver->bitdepth = bitdepth;
@@ -465,55 +487,85 @@ net_driver_new (jack_client_t * client,
         resample_factor_up = resample_factor;
 
     // Now open the socket, and wait for the first packet to arrive...
-    driver->sockfd = socket(PF_INET, SOCK_DGRAM, 0);
-    if (driver->sockfd == -1) {
-        jack_info("socket error");
+    driver->sockfd = socket (PF_INET, SOCK_DGRAM, 0);
+    if (driver->sockfd == -1)
+    {
+        jack_info ("socket error");
         return NULL;
     }
     address.sin_family = AF_INET;
     address.sin_port = htons(driver->listen_port);
     address.sin_addr.s_addr = htonl(INADDR_ANY);
-    if (bind(driver->sockfd, (struct sockaddr *) &address, sizeof(address)) < 0) {
+    if (bind (driver->sockfd, (struct sockaddr *) &address, sizeof (address)) < 0)
+    {
         jack_info("bind error");
         return NULL;
     }
 
-    driver->outsockfd = socket(PF_INET, SOCK_DGRAM, 0);
-    if (driver->sockfd == -1) {
-        jack_info("socket error");
+    driver->outsockfd = socket (PF_INET, SOCK_DGRAM, 0);
+    if (driver->sockfd == -1)
+    {
+        jack_info ("socket error");
         return NULL;
     }
     driver->srcaddress_valid = 0;
 
-    jacknet_packet_header *first_packet = alloca(sizeof(jacknet_packet_header));
-    socklen_t address_size = sizeof(struct sockaddr_in);
+    jacknet_packet_header *first_packet = alloca (sizeof (jacknet_packet_header));
+    socklen_t address_size = sizeof (struct sockaddr_in);
 
-    jack_info("Waiting for an incoming packet !!!");
-    jack_info("*** IMPORTANT *** Dont connect a client to jackd until the driver is attached to a clock source !!!");
+    jack_info ("Waiting for an incoming packet !!!");
+    jack_info ("*** IMPORTANT *** Dont connect a client to jackd until the driver is attached to a clock source !!!");
+
+    if (netjack_poll (driver->sockfd, 500))
+        first_pack_len = recvfrom (driver->sockfd, first_packet, sizeof (jacknet_packet_header), 0, (struct sockaddr*) & driver->syncsource_address, &address_size);
+    else
+        first_pack_len = 0;
 
-    int first_pack_len = recvfrom(driver->sockfd, first_packet, sizeof(jacknet_packet_header), 0, (struct sockaddr*) & driver->syncsource_address, &address_size);
     driver->srcaddress_valid = 1;
 
-    jack_info("first_pack_len=%d", first_pack_len);
-    // A packet is here.... If it wasnt the old trigger packet containing only 'x' evaluate the autoconf data...
-
     driver->mtu = 0;
 
-    if (first_pack_len == sizeof(jacknet_packet_header)) {
-        packet_header_ntoh(first_packet);
+    if (first_pack_len == sizeof (jacknet_packet_header))
+    {
+        packet_header_ntoh (first_packet);
 
-        jack_info("AutoConfig Override !!!");
-        if (driver->sample_rate != first_packet->sample_rate) {
-            jack_info("AutoConfig Override: sample_rate = %d", first_packet->sample_rate);
+        jack_info ("AutoConfig Override !!!");
+        if (driver->sample_rate != first_packet->sample_rate)
+        {
+            jack_info ("AutoConfig Override: Master JACK sample rate = %d", first_packet->sample_rate);
             driver->sample_rate = first_packet->sample_rate;
         }
 
-        if (driver->period_size != first_packet->period_size) {
-            jack_info("AutoConfig Override: period_size = %d", first_packet->period_size);
+        if (driver->period_size != first_packet->period_size)
+        {
+            jack_info ("AutoConfig Override: Master JACK period size is %d", first_packet->period_size);
             driver->period_size = first_packet->period_size;
         }
+        if (driver->capture_channels_audio != first_packet->capture_channels_audio)
+        {
+            jack_info ("AutoConfig Override: capture_channels_audio = %d", first_packet->capture_channels_audio);
+            driver->capture_channels_audio = first_packet->capture_channels_audio;
+        }
+        if (driver->capture_channels_midi != first_packet->capture_channels_midi)
+        {
+            jack_info ("AutoConfig Override: capture_channels_midi = %d", first_packet->capture_channels_midi);
+            driver->capture_channels_midi = first_packet->capture_channels_midi;
+        }
+        if (driver->playback_channels_audio != first_packet->playback_channels_audio)
+        {
+            jack_info ("AutoConfig Override: playback_channels_audio = %d", first_packet->playback_channels_audio);
+            driver->playback_channels_audio = first_packet->playback_channels_audio;
+        }
+        if (driver->playback_channels_midi != first_packet->playback_channels_midi)
+        {
+            jack_info ("AutoConfig Override: playback_channels_midi = %d", first_packet->playback_channels_midi);
+            driver->playback_channels_midi = first_packet->playback_channels_midi;
+        }
+        driver->capture_channels  = driver->capture_channels_audio + driver->capture_channels_midi;
+        driver->playback_channels = driver->playback_channels_audio + driver->playback_channels_midi;
 
         driver->mtu = first_packet->mtu;
+        jack_info ("MTU is set to %d bytes", first_packet->mtu);
         driver->latency = first_packet->latency;
     }
 
@@ -525,15 +577,17 @@ net_driver_new (jack_client_t * client,
     driver->net_period_down = (float) driver->period_size / (float) resample_factor;
     driver->net_period_up = (float) driver->period_size / (float) resample_factor_up;
 
-    driver->rx_buf = malloc(sizeof(jacknet_packet_header) + driver->net_period_down * driver->capture_channels * get_sample_size(driver->bitdepth));
+    /* TODO: this seems... useles */
+    rx_bufsize = sizeof (jacknet_packet_header) + driver->net_period_down * driver->capture_channels * get_sample_size (driver->bitdepth);
+    driver->rx_buf = malloc (rx_bufsize);
+    driver->pkt_buf = malloc (rx_bufsize);
+    global_packcache = packet_cache_new (driver->latency + 5, rx_bufsize, driver->mtu);
 
-    // XXX: dont need it when packet size < mtu
-    driver->pkt_buf = malloc(sizeof(jacknet_packet_header) + driver->net_period_down * driver->capture_channels * get_sample_size(driver->bitdepth));
-
-    int rx_bufsize = sizeof(jacknet_packet_header) + driver->net_period_down * driver->capture_channels * get_sample_size(driver->bitdepth);
-    global_packcache = packet_cache_new(driver->latency + 5, rx_bufsize, 1400);
-
-    jack_info("net_period: (up:%d/dn:%d)", driver->net_period_up, driver->net_period_down);
+    jack_info ("netjack: period   : up: %d / dn: %d", driver->net_period_up, driver->net_period_down);
+    jack_info ("netjack: framerate: %d", driver->sample_rate);
+    jack_info ("netjack: audio    : cap: %d / pbk: %d)", driver->capture_channels_audio, driver->playback_channels_audio);
+    jack_info ("netjack: midi     : cap: %d / pbk: %d)", driver->capture_channels_midi, driver->playback_channels_midi);
+    jack_info ("netjack: buffsize : rx: %d)", rx_bufsize);
     return (jack_driver_t *) driver;
 }
 
@@ -548,7 +602,7 @@ driver_get_descriptor ()
 
     desc = calloc (1, sizeof (jack_driver_desc_t));
     strcpy (desc->name, "net");
-    desc->nparams = 9;
+    desc->nparams = 11;
 
     params = calloc (desc->nparams, sizeof (jack_driver_param_desc_t));
 
@@ -568,6 +622,22 @@ driver_get_descriptor ()
     strcpy (params[i].short_desc, "Number of playback channels (defaults to 2)");
     strcpy (params[i].long_desc, params[i].short_desc);
 
+    i++;
+    strcpy (params[i].name, "midi inchannels");
+    params[i].character  = 'I';
+    params[i].type       = JackDriverParamUInt;
+    params[i].value.ui   = 1U;
+    strcpy (params[i].short_desc, "Number of midi capture channels (defaults to 1)");
+    strcpy (params[i].long_desc, params[i].short_desc);
+
+    i++;
+    strcpy (params[i].name, "midi outchannels");
+    params[i].character  = 'O';
+    params[i].type       = JackDriverParamUInt;
+    params[i].value.ui   = 1U;
+    strcpy (params[i].short_desc, "Number of midi playback channels (defaults to 1)");
+    strcpy (params[i].long_desc, params[i].short_desc);
+
     i++;
     strcpy (params[i].name, "rate");
     params[i].character  = 'r';
@@ -644,6 +714,8 @@ driver_initialize (jack_client_t *client, const JSList * params)
     jack_nframes_t period_size = 1024;
     unsigned int capture_ports = 2;
     unsigned int playback_ports = 2;
+    unsigned int capture_ports_midi = 1;
+    unsigned int playback_ports_midi = 1;
     unsigned int listen_port = 3000;
     unsigned int resample_factor_up = 0;
     unsigned int bitdepth = 0;
@@ -664,6 +736,14 @@ driver_initialize (jack_client_t *client, const JSList * params)
                 playback_ports = param->value.ui;
                 break;
 
+            case 'I':
+                capture_ports_midi = param->value.ui;
+                break;
+
+            case 'O':
+                playback_ports_midi = param->value.ui;
+                break;
+
             case 'r':
                 sample_rate = param->value.ui;
                 break;
@@ -694,8 +774,9 @@ driver_initialize (jack_client_t *client, const JSList * params)
         }
     }
 
-    return net_driver_new (client, "net_pcm", capture_ports,
-                           playback_ports, sample_rate, period_size,
+    return net_driver_new (client, "net_pcm", capture_ports, playback_ports,
+                           capture_ports_midi, playback_ports_midi,
+                           sample_rate, period_size,
                            listen_port, handle_transport_sync,
                            resample_factor, resample_factor_up, bitdepth);
 }
